﻿//-----------------------------------------------------------
// Chebyshev 補間による第１近似式の係数を求める
// （近似関数は sin(πx/2) で近似区間は -1 ～ 1 の場合）
//
//		作成者：三上直樹，2017/07/07 作成，(C)三上直樹 2017
//-----------------------------------------------------------

using System;
using System.Windows.Forms;

using Mikami.WindowsApplication;

namespace SinChebInt
{
	public partial class Form1 : Form
	{
		private const int MAX_ORDER_ = 20;
        private int order_;
		
		public Form1()
		{
			InitializeComponent();

			EnterToTab.Set(this);
		}

		private void button1_Click(object sender, EventArgs e)
		{
            double[] a = new double[MAX_ORDER_+1];
            double[] b = new double[MAX_ORDER_+1];
            double[] c = new double[MAX_ORDER_+1];
            double x1 =  -1.0;	// 近似区間の下限
            double x2 =  1.0;	// 近似区間の上限

			int order;
			if (!textBox2.ToInt(out order, v => v <= MAX_ORDER_,
										   "20 以下にしてください．")) return;
			if (order_ == order) return;

			order_ = order;
            ChebyshevInt(order_, Sinx, x1, x2, a);		// Chebyshev 多項式による近似式の係数
            PolyX(order_, a, b);						// 区間[-1, 1] の x に関する多項式の係数
            VariableTransform(order_, x1, x2, b, c);	// 区間[x1,x2] の x に関する多項式の係数

            textBox1.AppendText(order_.ToString() + " 次\r\n");

            for (int j=0; j<=order_; j++)
                textBox1.AppendText(String.Format(" c({0,2:D}) = {1,12:F8}\r\n",
												  j, c[j]));
            textBox1.AppendText("\r\n");
		}

		//------------------------------------------------------------------------
		// Chebyshev 補間
		//   order 次数
		//   func  近似すべき関数
		//   x1	   近似区間の下限
		//   x2	   近似区間の上限
		//   a	   得られる係数（Chebyshev 多項式による近似式の係数）
		private void ChebyshevInt(int order, Func<double, double> func,
								  double x1, double x2, double[] ak)
		{
			double[] fx = new double[order+1];
			double h = 0.5*Math.PI/(order+1);

			for	 (int k=0; k<=order; k++)
			{
				double rk = Math.Cos(Math.PI*(k+0.5)/(order+1));	// rk: 区間[-1, 1] の変数
				fx[k] = func(rk*0.5*(x2-x1) + 0.5*(x2+x1));
			}
			for (int n=0; n<=order; n++)
			{
				ak[n] = 0.0;
				for (int k=0; k<= order; k++)
					ak[n] = ak[n] + fx[k]*Math.Cos(n*h*(2*k+1));
				if (n==0) ak[n] = ak[n]/(order+1);
				else	  ak[n] = 2.0*ak[n]/(order+1);
			}
		}


		//------------------------------------------------------------------------
		// x に関する多項式への変換
		//   order 次数
		//   a	   Chebyshev の多項式による近似式の係数
		//   b	   x に関する多項式の係数
		private void PolyX(int order, double[] a, double[] b)
		{
		    double[,] tc = new double[order+1, order+1];
			ChebyshevPoly(order, tc);
			for (int n=0; n<=order; n++)
			{
				b[n] = 0.0;
				for (int k=0; k<=order; k++) b[n] = b[n] + tc[k, n]*a[k];
			}
		}

		//------------------------------------------------------------------------
		// Chebyshev 多項式の係数計算
		private void ChebyshevPoly(int order, double[,] tc)
		{
		    tc[0, 0] = 1.0;
		    for (int k=1; k<=order; k++) tc[0, k] = 0.0;
		    tc[1, 0] = 0.0;
		    tc[1, 1] = 1.0;
		    for (int k=2; k<=order; k++) tc[1, k] = 0.0;
		    for (int j=2; j<=order; j++)
			{
		        tc[j, 0] = -tc[j-2, 0];
				for (int k=1; k<=order; k++)
					tc[j, k] = 2*tc[j-1, k-1] - tc[j-2, k];
			}
		}

		//------------------------------------------------------------------------
		// 区間 [x1, x2] の近似式の係数の計算
		//   order 次数
		//   x1	   区間の下限
		//   x2	   区間の上限
		//   b     区間 [-1, 1] の近似式の係数
		//   c	   区間 [x1, x2] の近似式の係数
		private void VariableTransform(int order, double x1, double x2, double[] b, double[] c)
		{
		    double[,] binom = new double[order, order+1];
		    Binomial(order, binom);			// ２項係数の計算
		    double a1 = 1.0/(0.5*(x2-x1));
		    double a2 = -0.5*(x1+x2);
		    c[0] = b[0];
		    for (int n=1; n<=order; n++) c[n] = 0.0;
		    for (int n=1; n<=order; n++)
				for (int k=0; k<=n; k++)
					c[k] = c[k] + binom[n-1, k]*PowInt(a1,n)*PowInt(a2,n-k)*b[n];
		}

		//------------------------------------------------------------------------
		// ２項係数の計算
		//   C(j+1, k) = binom[j][k]
		private void Binomial(int order, double[,] binom)
		{
		    for (int j=0; j<order; j++) binom[j, 0] = 1.0;
		    binom[0, 1] = 1.0;
		    for (int k=2; k<=order; k++) binom[0, k] = 0.0;
		    for (int j=1; j<order; j++)
				for (int k=1; k<=order; k++)
					binom[j, k] = binom[j-1, k-1] + binom[j-1, k];
		}

		//------------------------------------------------------------------------
		// 近似する関数の定義
		private double Sinx(double x)
		{
		    return (Math.Sin(Math.PI*x/2.0));
		}

		//------------------------------------------------------------------------
		// n 乗（n: 0 以上の整数）
		private double PowInt(double x, int n)
		{
		    double tmp = 1.0;
		    for (int j=1; j<=n; j++) tmp = tmp*x;
			return (tmp);
		}
	}
}
