﻿//-----------------------------------------------------------------
// ミニマックス近似により近似多項式の係数を求める
//		（近似関数は sin(πx/2) で近似区間は -1 ～ 1 の場合）
//
//		このプログラムの近似多項式の係数の初期値は Chebyshev 補間に
//		より求めたものである．（プログラムは「SinChebyshev補間」）
//		を使用
//
//		作成者：三上直樹，2017/07/12 作成，(C)三上直樹 2017
//-----------------------------------------------------------------

using System;
using System.Windows.Forms;

using Mikami.Numeric;	// 三上作成のライブラリ．LU 分解により連立方程式を解く
						// ための LuDecomposition クラスが含まれる

namespace Sinミニマックス近似
{
	public partial class Form1 : Form
	{
		private const int MAX_ORDER_ = 20;	// このプログラムで扱えある最高の次数
		private struct ErrTbl{ public double x, y; };

		public Form1()
		{
			InitializeComponent();
		}

		private void button1_Click(object sender, EventArgs e)
		{
			double[] cn = new double[MAX_ORDER_+1];
			ErrTbl[] er = new ErrTbl[MAX_ORDER_+2];

			// ミニマックス近似を行うために必要なデータ
			// 以下は Chebyshev 補間により求めたデータ
			cn[0] =  0.0;           // x の 0 乗 の係数
			cn[1] =  1.57065736;    // x の 1 乗 の係数
			cn[2] =  0.0;           // x の 2 乗 の係数
			cn[3] = -0.64345777;    // x の 3 乗 の係数
			cn[4] =  0.0;           // x の 4 乗 の係数
			cn[5] =  0.07293465;    // x の 5 乗 の係数

			int order = 5;          // 近似多項式の次数
			int max_itr = 20;       // 繰り返しの上限
			double tol = 1.0e-10;	// 収束判定のための誤差の基準値
			double x1 = -1.0;       // 近似区間の下端
			double x2 =  1.0;       // 近似区間の上端
			// データはここまで

			textBox1.Clear();
			textBox1.AppendText("極値をとる x と，極値\r\n");
			int kExtr;
			ErrExtrm(cn, er, x1, x2, order, out kExtr);	// 誤差の極値を求める
			ErrShow(0, er, order, kExtr);				// 誤差の極値の表示

			double[] dcn = new double[MAX_ORDER_+2];
			double er_max = 0.0;
			double er_min = 1.0;
			for (int i=1; i<=max_itr; i++)
			{
				Solve(er, out dcn, order);				// 近似多項式の係数の補正値を計算

				for (int j=0; j<=order; j++)
					cn[j] = cn[j] + dcn[j];				// 係数の修正
				ErrExtrm(cn, er, x1, x2, order, out kExtr);	// 誤差の極値を求める
				ErrShow(i, er, order, kExtr);        	// 誤差の極値の表示

				// 極値の絶対値の最大値および最小値を求める
				er_max = 0.0;
				er_min = 1.0;
				for (int j=0; j<order+2; j++)
				{
					er_max = Math.Max(Math.Abs(er[j].y), er_max);
					er_min = Math.Min(Math.Abs(er[j].y), er_min);
				}
				// 極値の絶対値の最大値と最小値の差が基準値（tol）未満であれば終了
				if ( (er_max-er_min) < tol ) break;
			}
			if ( (er_max - er_min) > tol ) textBox1.AppendText("収束が不十分です");
			textBox1.AppendText("\r\n誤差の絶対値の最大値 ＝ " + er_max.ToString("E8") + "\r\n");
			textBox1.AppendText("得られた係数：\r\n");
			for (int j=0; j<=order; j++)            // 得られた係数の表示
				textBox1.AppendText(String.Format("  c{0,1:D} = {1,16:F12}\r\n",
									j, cn[j]));
		}

		// 誤差の極値を求める
		//		k: 極値の数（区間の両端を含む数）
		private void ErrExtrm(double[] cn, ErrTbl[] er, double x1, double x2, int order, out int k)
		{
			double dx = 1.0/32768.0;
			double ym1 = ErrEval(cn, x1-dx, order);
			double y0 = ErrEval(cn, x1, order);
			er[0].x = x1;
			er[0].y = y0;
			 
			k = 1;
			for (double x=x1; x<=x2; x=x+dx)
			{
				double yp1 = ErrEval(cn, x+dx, order);
				if ( ((y0 >= ym1) && (y0 > yp1)) || ((y0 <= ym1) && (y0 < yp1)) )
				{
					er[k].x = x;    // 極値をとる変数の値
					er[k].y = y0;   // 極値
					k++;
				}
				ym1 = y0;
				y0  = yp1;
			}
			er[k].x = x2;
			er[k].y = y0;
			if ((k+1) < (order+2))  // 極値の数が 次数+2 に満たない場合
			{
				MessageBox.Show("第１近似式が不適切なため誤差の極値の数が不足です，極値数： "
							   +  (k+1).ToString());
				Environment.Exit(0);
			}
			k++;	// k は区間の両端を含めた個数
		}

		// 誤差の極値の表示
		private void ErrShow(int itr, ErrTbl[] er, int order, int k)
		{
			if (itr == 0) textBox1.AppendText("第１近似式\r\n");
			else
				textBox1.AppendText(itr.ToString() + "回目\r\n");
			for (int j=0; j<k; j++)     // 極値の表示
				textBox1.AppendText(String.Format("  x = {0,12:F8}, {1,16:E8}\r\n",
												  er[j].x, er[j].y));
		}

		// 近似誤差の計算
		private double ErrEval(double[] cn, double x, int order)
		{
			// 近似関数に応じて書換える
			return ( Math.Sin(0.5*Math.PI*x) - Poly(x, order, cn) );
		}

		// 近似多項式の係数の補正値を求める
		// （連立方程式の解法は LU 分解を利用）
		private void Solve(ErrTbl[] er, out double[] dcn, int order)
		{
			double[,] a = new double[order+2, order+2];
			double[] b = new double[order+2];

			for (int j=0; j<order+2; j++)
			{
				a[j, 0] = 1;
				for (int k=1; k<=order; k++) a[j, k] = Math.Pow(er[j].x, k);
				a[j, order+1] = Math.Pow(-1.0, j);
				b[j] = er[j].y;
			}

			// LuDecomposition: 三上の作成した，LU 分解で連立方程式を解くためのクラス
			LuDecomposition mx = new LuDecomposition(a);

			if (mx.Det == null)
			{
				MessageBox.Show("連立方程式が解けないので終了します");
				Environment.Exit(0);
			}

			dcn = mx.Solve(b);	// 連立方程式を解いた結果を取り出す
		}

		// 多項式の値の計算，ホーナー法を利用
		//		y = cn[0] + cn[1]*x + cn[2]*x*X + ･････
		private double Poly(double x, int order, double[] cn)
		{
			double y = cn[order];
			for (int n=order-1; n>=0; n--)
				y = y*x + cn[n];
			return y;
		}
	}
}
